/*
 * list.h - Dynamic doubly-linked lists library
 * Copyright (C) 1994-2000  Matthew Mondor
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#ifndef LIST_CH


#define LIST_CH


// INCLUDES
// ========

#include <stdlib.h>
#ifndef MMTYPES_H
#include "mmtypes.h"
#endif



// STRUCTURES
// ==========

typedef struct lnode
{
  struct lnode *ln_next;
  struct lnode *ln_prev;
}
lnode;

typedef struct list
{
  LONG l_lnodelen;
  LONG l_lnodes;
  lnode *l_toplnode;
  lnode *l_bottomlnode;
  lnode *l_currentlnode;
}
list;



// PROTOTYPES
// ==========

list *createlist (LONG);
list *freelist (list *);
lnode *appendnode (list *);
lnode *deletenode (list *);
lnode *nextnode (list *);
lnode *topnode (list *);
LONG totalnodes (list *);
VOID makecurrent (list *, lnode *);



// FUNCTIONS
// =========

/* Creates a list in memory, with no lnodes in it, returns null on error
 * This is necessary to create a list.
 */
list *
createlist (LONG nodelen)
{
  list *lst;

  if (nodelen)
    {
      if ((lst = (list *) calloc (1, sizeof (list))))
	{
	  lst->l_lnodelen = nodelen;
	  return (lst);
	}
    }
  return (NULL);
}


// Frees a list from memory
list *
freelist (list * lst)
{
  lnode *node, *tmp;

  if (lst)
    {
      node = lst->l_toplnode;
      if (node)
	{
	  // Free the whole list
	  while (node)
	    {
	      tmp = node->ln_next;
	      free (node);
	      node = tmp;
	    }
	}
      free (lst);
    }

  return (NULL);
}


/* Appends an empty lnode to the list, at the end of all other
 * lnodes. The lnode can be initialised with values afterwards.
 * The new lnode is the new current lnode of the list.
 */
lnode *
appendnode (list * lst)
{
  lnode *node, *tmp;

  if (lst)
    {
      if ((node = (lnode *) calloc (1, lst->l_lnodelen)))
	{

	  // Link newly allocated lnode to the last lnode in the list
	  if ((tmp = lst->l_bottomlnode))
	    {
	      // There are other lnode(s), link it to last and make last
	      tmp->ln_next = node;
	      node->ln_prev = tmp;
	      lst->l_bottomlnode = node;
	    }
	  else
	    {
	      // Only lnode, so only link to list
	      lst->l_bottomlnode = lst->l_toplnode = node;
	    }

	  // Increase number of lnodes
	  lst->l_lnodes++;

	  // Make new lnode current
	  lst->l_currentlnode = node;
	  return (node);

	}
    }

  return (NULL);
}


/* Deletes the lnode at the current position in the list file.
 * The returned pointer is the updated current lnode position, pointing
 * at the same logical number: the lnode just after the one we deleted.
 */
lnode *
deletenode (list * lst)
{
  lnode *node, *prev, *next;

  if (lst)
    {
      if ((node = lst->l_currentlnode))
	{

	  // Remember link pointers of current lnode
	  prev = node->ln_prev;
	  next = node->ln_next;
	  // Free current lnode
	  free (node);
	  // Decrease lnode counter
	  lst->l_lnodes--;

	  // Re-link list/nextlnode pointers
	  if (prev)
	    {
	      /* There are lnodes before.
	       * If there are lnodes after, link to last lnode
	       */
	      if (next)
		next->ln_prev = prev;
	      else
		{
		  // There are no lnodes after
		  prev->ln_next = NULL;
		  lst->l_bottomlnode = prev;
		}
	    }
	  else
	    {
	      // There are no lnodes before
	      if (next)
		{
		  next->ln_prev = NULL;
		  lst->l_toplnode = next;
		}
	      else
		lst->l_toplnode = NULL;
	    }
	  if (next)
	    {
	      /* There are lnodes after.
	       * If there are lnodes before, link to next lnode
	       */
	      if (prev)
		prev->ln_next = next;
	      else
		{
		  // There are no lnodes before
		  next->ln_prev = NULL;
		  lst->l_toplnode = next;
		}
	    }
	  else
	    {
	      // There are no lnodes after
	      if (prev)
		{
		  prev->ln_next = NULL;
		  lst->l_bottomlnode = prev;
		}
	      else
		lst->l_bottomlnode = NULL;
	    }

	  // Set current lnode to next lnode (may be null)
	  lst->l_currentlnode = next;

	  return (next);
	}
    }
  return (NULL);
}


// Increases the current pointer to next lnode and returns pointer
lnode *
nextnode (list * lst)
{
  lnode *node;

  if (lst)
    {
      if ((node = lst->l_currentlnode))
	node = node->ln_next;
      if (node)
	{
	  lst->l_currentlnode = node;
	  return (node);
	}
    }
  return (NULL);
}


// Jumps to first lnode in list, returns new lnode pointer
lnode *
topnode (list * lst)
{
  if (lst)
    {
      lst->l_currentlnode = lst->l_toplnode;
      return (lst->l_currentlnode);
    }
  return (NULL);
}


// Returns the total number of lnodes into a list
LONG
totalnodes (list * lst)
{
  if (lst)
    {
      return (lst->l_lnodes);
    }
  return (0);
}


/* Lets the programmer to decide which lnode must be current.
 * Beware! It must be a valid lnode of the list specified!
 */
VOID
makecurrent (list * lst, lnode * node)
{
  if (lst && node)
    {
      lst->l_currentlnode = node;
    }
}



#endif
